# options.py

'''
TODO
 - ajouter une option permettant de faire un fasta d'une release de transcriptome
'''

"""
Options and arguments to bo passed on to the program.
"""

import os
import sys
import argparse
import info
import tempfile
import subprocess

from color import *


class EditConfig(argparse.Action):
    """
    https://stackoverflow.com/questions/8632354/python-argparse-custom-actions-with-additional-arguments-passed
    https://stackoverflow.com/questions/11001678/argparse-custom-action-with-no-argument
    """
    def __init__(self, option_strings, conf, *args, **kwargs):
        """ Function doc """
        super(EditConfig, self).__init__(option_strings=option_strings, *args, **kwargs)
        self.conf = conf

    def __call__(self, parser, namespace, values, option_strings):
        self.conf.edit()
        sys.exit()


def usage(conf):
    """
        Help function with argument parser.
    """
    parser = argparse.ArgumentParser(
        description=info.DOC,
        epilog = info.EXAMPLES,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    exclusive = parser.add_mutually_exclusive_group(required=True)
    exclusive.add_argument('-s', '--selection',
                        help=(
                            "list of gene IDs (ENSG, gene Symbol or alias) or transcript IDs "
                            "(ENST) from which you want to extract specific kmers from. For genes, "
                            "kmerator search specific kmers along the gene. For transcripts, it "
                            "search specific kmers to the transcript. You can also give a file "
                            "with yours genes/transcripts separated by space, tab or newline. "
                            "If you want to use your own unannotated sequences, you "
                            "must give your fasta file with --fasta-file option."
                            ),
                        nargs='+',
                        )
    exclusive.add_argument('-f', '--fasta-file',
                        help=(
                            "Use this option when yours sequences are unannonated or provided "
                            "by a annotation file external from Ensembl. Otherwise, "
                            "use --selection option."
                            ),
                        )
    parser.add_argument('-d', '--datadir',
                        help=("Storage directory for kmerator datasets."
                              "We recommend to set this parameter by editing the configuration "
                              f"file ({info.APPNAME} --edit)"
                              ),
                        )
    parser.add_argument('-g', '--genome',
                        help=(
                            "Genome jellyfish index (.jf) to use for k-mers requests."
                            ),
                        required=True,
                        )
    parser.add_argument('-S', '--specie',  ### replace -a --appris in julia version
                        help=(
                            "indicate a specie referenced in Ensembl, to help, follow the link "
                            "https://rest.ensembl.org/documentation/info/species. You can use "
                            "the 'name', the 'display_name' or any 'alias'. For example human or "
                            "homo_sapiens are valid (default: human)."
                            ),
                        default='human',
                        )
    parser.add_argument('-k', '--kmer-length',
                        type=int,
                        help="k-mer length that you want to use (default 31).",
                        default=31,
                        )
    parser.add_argument('-r', '--release',
                        help="release of transcriptome (default: last).",
                        default="last",
                        )
    # ~ parser.add_argument('--chimera',
                        # ~ action='store_true',
                        # ~ help="Only with '-f/--fasta-file' option.",
                        # ~ )
    parser.add_argument('--stringent',
                        action='store_true',
                        help=(
                            "Only for genes with '-s/--selection' option: use this option if you want "
                            "to select gene-specific k-mers present in ALL transcripts for your "
                            "gene. If false, a k-mer is considered as gene-specific if present "
                            "in at least one isoform of your gene of interest."
                            ),
                        )
    parser.add_argument('-T', '--max-on-transcriptome',
                        type=int,
                        help=("Only for genes with '-f/--fasta-file' option: with unanotated data, "
                            "specific kmers are not supposed to be found in the transcriptome, you "
                            "can change this behavior for special cases (default: 0)"
                            ),
                        default=0,
                        )
    parser.add_argument('-G', '--max-on-genome',
                        type=int,
                        help=(
                            "Only for genes with '-f/--fasta-file' option: typically, specific "
                            "kmers are not supposed to be found more than once in the genome, you "
                            "can change this behavior for special cases, like chimera (default: 1)"
                            ),
                        default=1 if any(set(['-f', '--fasta-file']) & set(sys.argv)) else None,
                        )
    parser.add_argument('-o', '--output',
                        help="output directory, created if not exists (default: 'output')",
                        default='output',
                        )
    parser.add_argument('-t', '--thread',
                        type=int,
                        help="run n process simultaneously (default: 1)",
                        default=1,
                        )
    parser.add_argument('--tmpdir',
                        help="directory to temporary file (default: /tmp/kmerator_<random>)",
                        default=None,
                        )
    parser.add_argument('-D', '--debug',
                        action='store_true',
                        help="Show more details while Kmerator is running.",
                        )
    parser.add_argument('--keep',
                        action='store_true',
                        help=("keep intermediate files (sequences, indexes, separate kmers and "
                            "contigs files)."
                            ),
                        )
    parser.add_argument('-y', '--yes',
                        action='store_true',
                        help=("assumes 'yes' as the prompt answer, run non-interactively."),
                        )
    parser.add_argument('-e', '--edit-config',
                        action=EditConfig,
                        nargs=False,
                        conf=conf,
                        help='Edit config file',
                        )
    exclusive.add_argument('-l', '--list-dataset', '--list-datasets',
                        help=("list the local datasets (based on the datadir option)."),
                        action='store_true',
                        )
    exclusive.add_argument('--rm-dataset',
                        action='store_true',
                        help="remove a dataset, according with --specie and --release options",
                      )
    exclusive.add_argument('--mk-dataset',
                        action='store_true',
                        help="make a dataset, according with --specie and --release options",
                      )
    exclusive.add_argument('--last-avail', '--last-available',
                        action='store_true',
                        help="last release available on Ensembl",
                       )
    exclusive.add_argument('-u', '--update-dataset',
                        action="store_true",
                        help="builds a new dataset if a new version is found on Ensembl",
                       )
    exclusive.add_argument('--info',
                        nargs="+",
                        metavar="gene",
                        help=("get some information about gene. Multiple entries are allowed or "
                              "a text file with list of genes"),
                        )
    parser.add_argument('-a', '--all',
                        action='store_true',
                        help=("only with '--info' option. Give more info, like transcript sequences"),
                        )
    parser.add_argument('-v', '--version',
                        action='version',
                        version=f'{parser.prog} v{info.VERSION}',
                        )

    ### Set default values define by Config class (using ConfigParser)
    default_conf = conf.content['CMD_ARGS']
    parser.set_defaults(**default_conf)
    ### Reset `required` attribute when provided from config file
    for action in parser._actions:
        if action.dest in default_conf:
            action.required = False
            ### handle booleans cause ConfigParse return a string.
            if action.__class__.__name__ == "_StoreTrueAction":
                action.default = True if action.default.lower() in ['true', 'yes', '1', 'on'] else False

    ### Go to 'usage()' without arguments or stdin
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    args = parser.parse_args()

    ### Check args
    checkup_args(args)

    return args


def checkup_args(args):
    """ checkup some options. """
    ### --datadir - at the first use, configure path of kmerator files
    if not args.datadir:
        mesg = ("\nAt the first use of kmerator, you need to configure a path for the files that "
               "kmerator will store: a fasta file (fa), a jellyfish file (jf) and a binary"
               " file (pkl) per release. Each release takes more than 2GB "
               "(2.2GB for release 108), so allow enough space.\n\n"
               f"Also, to avoid to set the '--genome' options each time you launch {info.APPNAME},"
               " you should specify the path of the jellyfish index of your reference genome.\n\n"
               "Note that other parameters are available in the configuration file, so that you "
               "don't have to enter some recurring options (such as the species "
               "for example).\n\n"
               f"Please, use the {BOLD}{DARKCYAN}--edit-config{ENDCOL} "
               f"(or {BOLD}{DARKCYAN}-e{ENDCOL}) option first to set the 'datadir' path parameter.\n")
        sys.exit(mesg)
    ### --datadir - check if path exists and right read is set
    if not os.path.isdir(args.datadir):
        sys.exit(f"{ERROR}Error: {args.datadir!r} not found or not a directory.{ENDCOL}")
    if not os.access(args.datadir, os.R_OK) or not os.access(args.datadir, os.X_OK):
        sys.exit(f"{ERROR}Error: insufficient rights in {args.datadir!r}.{ENDCOL}")
    ### --fasta-file - check if query fasta file is present
    if args.fasta_file and not os.path.isfile(args.fasta_file):
        sys.exit(f"{ERROR}Error: {args.fasta_file!r} not found.{ENDCOL}")
    ### --fasta-file - Check query fasta file extension
    if args.fasta_file and args.fasta_file.split('.')[-1] not in ("fa", "fasta"):
        sys.exit(f" {ERROR}Error: {os.path.basename(args.fasta_file)} " \
                f"Does not appears to be a fasta file.{ENDCOL}")
    if args.fasta_file:
        check_fasta_file(args)
    ### --selection - Check Gene Select option
    if args.selection:
        for gene in args.selection:
            if gene.startswith('ENS') and '.' in gene:
                sys.exit(f"{ERROR}ENSEMBL annotations with version (point) like "
                        f"ENSTXXXXXXXX.1 is forbidden, just remove the '.1'."
                        )
        ### Define genes/transcripts provided when they are in a file
        if len(args.selection) == 1 and os.path.isfile(args.selection[0]):
            with open(args.selection[0]) as fh:
                args.selection = []
                for line in fh:
                    args.selection += line.split('#')[0].split()

    ### args.info: define genes/transcripts provided when they are in a file
    if args.info and len(args.info) == 1 and os.path.isfile(args.info[0]):
            with open(args.info[0]) as fh:
                args.info = []
                for line in fh:
                    args.info += line.split('#')[0].split()


    ### --genome - check jellifish genome
    if not os.path.isfile(args.genome):
        sys.exit(f"{ERROR}Error: file not found: {args.genome!r}.{ENDCOL}")
    cmd = f"jellyfish info {args.genome}"
    stdin = subprocess.check_output(cmd, shell=True).decode().rstrip().split('\n')
    stdin = { i[0]:i[1].strip() for i in [raw.split(':') for raw in stdin]}
    if not stdin['command']:            # not a jellyfish index
        sys.exit(f"{ERROR}Error: {args.genome!r} does not appear to be a jellyfish index.\n")
    if stdin['canonical'] == 'no':     # jellyfish index built without --canonical option
        sys.exit(f"{ERROR}Error: the genome jellyfish index must be built with --canonical option.\n"
                 f"{ENDCOL}From kmerator v0.9.0, it must be built with the '--canonical' option.\n"
                 "Example to built a jellyfish index:\n"
                 f"{YELLOW}jellyfish count /genomes/GRCh38.fa -m 31 -s 100M -t 8 --canonical "
                 "-o GRCh38.jf")
    ### --chimera level works only with --fasta-file option
    # ~ if args.chimera and not args.fasta_file:
        # ~ sys.exit(f"{ERROR}Error: '--chimera' needs '-f/--fasta-file' option.{ENDCOL}")
    ### --max-on-genome level works only with --fasta-file option
    if args.max_on_genome and not args.fasta_file:
        sys.exit(f"{ERROR}Error: '--max-on-genome' needs '-f/--fasta-file' option.{ENDCOL}")
    ### --max-on-transcriptome level works only with --fasta-file option
    if args.max_on_transcriptome and not args.fasta_file:
        sys.exit(f"{ERROR}Error: '--max-on-transcriptome' needs '-f/--fasta-file' option.{ENDCOL}")
    ### temporary directory
    if not args.keep:
        if not args.tmpdir:
            args.tmpdir = tempfile.mkdtemp(prefix="kmerator_")
        elif not os.path.isdir(args.tmpdir):
            sys.exit(f"{ERROR}Error: temporary directory not found ({args.tmpdir}).{ENDCOL}")
        else:
            args.tmpdir = tempfile.mkdtemp(prefix="kmerator_", dir=args.tmpdir)
    else:
        args.tmpdir = args.output


def check_fasta_file(args):
    headers = set()
    with open(args.fasta_file) as fh:
        nb = 1
        for row in fh:
            if row.startswith('>'):
                header = row.split(' ')[0]
                if header in headers:
                    sys.exit(f"{ERROR}ErrorFastaFile: sequence identifier must be unique.\n"
                                    f"    {header.rstrip()!r} is not unique (line {nb}).")
                headers.add(header)
            nb += 1
    del headers



#!/usr/bin/env python3

"""
Decomposition of transcript or gene sequences and extraction of their specific k-mers
Once:
  - Download some file of Ensembl database, regarding of requested release of transcriptome
    (once per release) to build files for kmerator
Per request:
  1. Load kmerator files of the release resquested
  3. Build sequences for each transcript
  4. Get specific kmers/contigs
  5. Concatene specific kmers/contigs files
  6. write and show report
  7. exit removing temporary file unless the --keep option is set
"""

import sys
import os
import signal
from functools import partial
import shutil
from datetime import datetime
import getpass
import pickle

### local packages
import info
from config import Config
from dataset import Dataset
from kmerize import SpecificKmers
import geneinfo
from options import usage
from color import *
import exit



def main():
    ### Handle arguments
    conf = Config(info.APPNAME)
    args = usage(conf)
    if args.debug: print(f"{DEBUG}Args:\n {args}\n{ENDCOL}")

    ### Stops running indexes on exit (Ctrl C, kill -15)
    global sigint_handler
    sigint_handler = partial(sigint_handler, args=args)
    signal.signal(signal.SIGINT, sigint_handler)

    ### check files of the dataset (download if needed)
    dataset = Dataset(args)

    ### Handle dataset options: '--list-dataset', '--rm-dataset'
    if args.list_dataset: dataset.list()
    if args.rm_dataset: dataset.remove()
    if args.mk_dataset: dataset.make()
    if args.last_avail: dataset.last_available()
    if args.update_dataset: dataset.update_last()
    if args.info: geneinfo.info(args)


    ### load kmerator geneinfo file as dict
    print(f" 🧬 Load dataset {args.release!r}.")

    ### Important objects
    items = []  # {given: xxx, ENST: xxx, ENSG: xxx, symbol: xxx, type: xxx, f_id: xxx, seq: xxx}

    report = {'failed': [], 'done': [], 'multiple': [], 'warning': []}
    transcriptome_dict = None       # not used with --fasta-file option
    geneinfo_dict = None            # not used with '--fasta-file' option

    ### load transcriptome if --selection option is set
    if args.selection:
        ### create shared objects among multiple processes with Manager()
        transcriptome_dict = dataset.load_transcriptome()

    ### load geneinfo
    geneinfo_dict = dataset.load_geneinfo()
    ### Find transcripts according the selection
    items = find_items(args, report, geneinfo_dict)

    ### get specific kmers (using multithreading)
    print(f" 🧬 Extract specific kmers, please wait..")
    SpecificKmers(args, report, items, transcriptome_dict, geneinfo_dict)

    ### Concatene results
    print(f" 🧬 Build finals results and report")
    merged_results(args)

    ### show some final info in the prompt
    show_report(args, report)

    ### set markdown report
    markdown_report(args, report)

    ### Bye
    exit.gracefully(args)


def find_items(args, report, geneinfo_dict=None):
    """
    for each name given (symbol/alias/ENSG/ENST), get information from geneinfo_dict
    """
    items = []
    ### With '--selection' option
    if args.selection:
        for given in args.selection:
            found_count = 0
            found_transcripts = []
            item = given.upper()

            ### sometimes, canonical transcript is not define (old releases), get longest transcript
            # ~ canonical = lambda args, ENSG: (geneinfo_dict['gene'][ENSG]['canonical']
                                 # ~ if 'canonical' in geneinfo_dict['gene'][ENSG]
                                 # ~ else longest_transcript(args, ENSG))

            if item in geneinfo_dict['symbol']:
                ENSGs = geneinfo_dict['symbol'][item]
                for ENSG in ENSGs:
                    found_count += 1
                    found_transcripts.append(geneinfo_dict['gene'][ENSG]['canonical'])
                    items.append({'given': given, 'ENSG': ENSG, 'type': 'symbol',
                                       'ENST':   geneinfo_dict['gene'][ENSG]['canonical'],
                                       'symbol': geneinfo_dict['gene'][ENSG]['symbol'],
                                       })
            elif item in geneinfo_dict['alias']:
                ENSGs = geneinfo_dict['alias'][item]
                for ENSG in ENSGs:
                    found_count += 1
                    found_transcripts.append(geneinfo_dict['gene'][ENSG]['canonical'])
                    items.append({'given':  given, 'ENSG': ENSG, 'type': 'alias',
                                        'ENST':   geneinfo_dict['gene'][ENSG]['canonical'],
                                        'symbol': geneinfo_dict['gene'][ENSG]['symbol'],
                                        })
            elif item in geneinfo_dict['gene']:
                found_count += 1
                ENSG = item
                items.append({'given':  given, 'ENSG': ENSG, 'type': 'gene',
                                    'ENST':   geneinfo_dict['gene'][ENSG]['canonical'],
                                    'symbol': geneinfo_dict['gene'][ENSG].get('symbol', 'N/A'),
                                    })
            elif item in geneinfo_dict['transcript']:
                found_count += 1
                ENSG = geneinfo_dict['transcript'][item]
                items.append({'given':  given, 'ENST': item, 'type': 'transcript',
                                    'ENSG':   ENSG,
                                    'symbol': geneinfo_dict['gene'][ENSG].get('symbol', 'N/A'),
                                    })
            else:
                report['failed'].append(f"{given}: not found in transcriptome")

            if found_count > 1:
                report['multiple'].append(f"{given}: {found_count} ({', '.join([i for i in found_transcripts])})")


    ### With '--fasta-file' option
    if args.fasta_file:
        type = 'transcript'
        with open(args.fasta_file) as fh:
            seq = ""
            f_id = fh.readline()[1:].split(' ')[0].rstrip()
            for raw in fh:
                if raw.startswith('>'):
                    if len(seq) >= args.kmer_length:
                        items.append({'f_id': f_id, 'seq': seq, 'type': type})
                    else:
                        report['failed'].append(f"{f_id}: sequence to short ({len(seq)} < {args.kmer_length})")
                    f_id = raw[1:].split(' ')[0].rstrip()
                    seq = ""
                else:
                    seq += raw.rstrip()
            ### last f_id/sequence
            if len(seq) >= args.kmer_length:
                items.append({'f_id': f_id, 'seq': seq, 'type': type})
            else:
                report['failed'].append(f"{f_id}: sequence to short ({len(seq)} < {args.kmer_length})")

    return items


def longest_transcript(args, ENSG):
    print(args)
    print(ENSG)
    if args.debug: print(f"{YELLOW}Finding the longest transcript for the gene {ENSG}.{ENDCOL}")
    sys.exit(f"EOF (Dev)")
    transcripts_dict = { k:len(v) for (k,v) in transcriptome_dict.items() if k.startswith(f"{gene_name}:")}
    # ~ print(*[k for k in variants_dict], sep='\n')
    nb_transcripts = len(transcript_dict)
    if args.verbose: print(f"{Color.YELLOW}Number of variants: {nb_transcripts}")
    longest_transcript = None
    length = 0
    for k,v in transcript_dict.items():
        if v > length:
            length = v
            longest_transcript = ':'.join(k.split(':')[1:2])
    return longest_transcript


def merged_results(args):
    if not os.path.isdir(os.path.join(args.tmpdir, 'kmers')):
        return None
    for item in ['kmers', 'contigs', 'masked']:
        try:
            files = os.listdir(os.path.join(args.tmpdir, item))
        except FileNotFoundError:
            return
        if files:
            merged_file = os.path.join(args.output, f"{item}.fa")
            os.makedirs(args.output, exist_ok=True)
            with open(merged_file,'wb') as mergefd:
                for file in files:
                    with open(os.path.join(args.tmpdir, item, file),'rb') as fd:
                        shutil.copyfileobj(fd, mergefd)


def show_report(args, report):
    ### show some final info in the prompt
    if report['done']:
        print(f"{CYAN}\n Done ({len(report['done'])}):")
        for i,mesg in enumerate(report['done']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")

    if report['multiple']:
        print(f"{BLUE}\n Multiple responses ({len(report['multiple'])}):")
        for i,mesg in enumerate(report['multiple']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")
            # ~ for k,v in mesg.items():
                # ~ print(f"  - {k}: {' '.join(v)}")

    if report['failed']:
        print(f"{PURPLE}\n Failed ({len(report['failed'])}):")
        for i,mesg in enumerate(report['failed']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")

    if report['warning']:
        print(f"{RED}\n Warning ({len(report['warning'])}):")
        for i,mesg in enumerate(report['warning']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")

    print(f"{ENDCOL}")


def markdown_report(args, report):
    sel_or_fa = 'selection' if args.selection else 'fasta_file'
    to_report = [
        sel_or_fa, 'datadir', 'genome', 'specie', 'kmer_length', 'release', 'stringent',
        'max_on_transcriptome', 'max_on_genome', 'output', 'thread', 'keep', 'assembly',
    ]
    os.makedirs(args.output, exist_ok=True)
    with open(os.path.join(args.output, 'report.md'), 'w') as fh:
        fh.write('# kmerator report\n')
        fh.write(f"*date: {datetime.now().strftime('%Y-%m-%d %H:%M')}*  \n")
        fh.write(f'*login: {getpass.getuser()}*\n\n')
        fh.write(f"**kmerator version:** {info.VERSION}\n\n")
        ### report command line and args, included defaults args
        cmd_args = ''
        for k,v in vars(args).items():
            if k in to_report:
                k = k.replace('_', '-')
                if isinstance(v, list):
                    v = ' '.join(v)
                cmd_args += f" \\\n  --{k} {v}"
        # ~ cmd_args = ' \\\n  '.join([f"--{k} {v}" for k,v in vars(args).items() if v])
        # ~ command = f"command: \n{__appname__}{cmd_args}"
        command = f"{info.APPNAME}{cmd_args}"
        # ~ command = ' '.join(sys.argv).replace(' -', ' \\\n  -')
        fh.write(f"**Command:**\n\n```\n{command}\n```\n\n")
        fh.write(f"**Working directory:** `{os.getcwd()}`\n\n")
        fh.write(f"**Specie:** `{args.specie}`\n\n")
        fh.write(f"**Assembly:** `{args.assembly}`\n\n")
        fh.write(f"**Transcriptome release:** `{args.release}`\n\n")
        if report['done']:
            fh.write(f"**Genes/transcripts succesfully done ({len(report['done'])})**\n\n")
            for mesg in report['done']:
                fh.write(f"- {mesg}\n")
        if report['multiple']:
            fh.write(f"\n**Multiple Genes returned for one given by Ensembl API ({len(report['multiple'])})**\n\n")
            for mesg in report['multiple']:
                fh.write(f"- {mesg}\n")
                # ~ for k,v in mesg.items():
                    # ~ fh.write(f"- {k}: {' '.join(v)}\n")
        if report['failed']:
            fh.write(f"\n\n**Genes/transcripts missing ({len(report['failed'])})**\n\n")
            for mesg in report['failed']:
                fh.write(f"- {mesg}\n")
        if report['warning']:
            fh.write(f"\n\n**Warnings ({len(report['warning'])})**\n\n")
            for mesg in report['warning']:
                fh.write(f"- {mesg}\n")


def html_report(args, report):
    '''TODO (plus option to change report format)'''
    print("Work in progress...")


def yaml_report(args, report):
    '''TODO'''
    print("Work in progress...")


def json_report(args, report):
    '''TODO'''
    print("Work in progress...")


def sigint_handler(signal, frame, args):
    exit.gracefully(args)


if __name__ == '__main__':
    main()




import os
import sys
import requests
from bs4 import BeautifulSoup
from urllib.request import urlretrieve
import gzip
import pickle
import shutil
import tempfile
import threading

from color import *


class TranscriptomeBuilder:
    """
    - download 'cdna' and 'ncrna' from Ensembl ftp site
    - remove transcripts not located in regular chromosomes (avoid duplicates)
    - concatene a temporary file with results
    - build a jellyfish index
    - delete temporary files
    """

    def __init__(self, args, base_url, transcriptome_jf, report, chr):
        """ Class initialiser """
        self.args = args
        self.base_url = base_url
        self.report = report
        self.chr = chr
        self.transcript_count = 0

        if self.args.debug: print(f"{DEBUG}Build transcriptome, please wait...{ENDCOL}")
        ### select fasta file to download
        release = self.args.release

        self.tmpdir = tempfile.mkdtemp(prefix="kmerator_")
        self.downloaded_fasta = []
        self.temp_fasta_files = []

        if self.args.debug: print(f"{DEBUG}Download transcriptome files (release {release}), please wait...{ENDCOL}")
        ### download fasta gzipped files and clean them
        for item in ('cdna', 'ncrna'):
            self.build_fasta(item)

        ### concatene filtered cDNA and ncRNA (and remove temp fasta files)
        transcriptome_fa = self.mk_transcriptome()

        ### make index of transcriptome (kmc or jellyfish)
        self.mk_index(transcriptome_fa, transcriptome_jf)

        ### create a pickle file of transcriptome
        self.mk_pickle(transcriptome_fa)

        '''
        ### make index of transcriptome (kmc or jellyfish)
        th1_1 = threading.Thread(target=self.mk_index,
                               args=(transcriptome_fa, transcriptome_jf))
        ### create a pickle file of transcriptome
        th1_2 = threading.Thread(target=self.mk_pickle,
                               args=(transcriptome_jf,))
        th1_1.start()
        th1_2.start()
        th1_1.join()
        th1_2.join()
        '''

        ### Add to report
        self._to_report()

        ### remove intermediate files
        shutil.rmtree(self.tmpdir, ignore_errors=True)
        if not args.keep:
            os.remove(transcriptome_fa)


    def build_fasta(self, item):
        """
        for cdna and ncrna:
          - define url
          - download file
          - clean file
        """
        ### get url links for cDNA and ncRNA fasta files
        url = os.path.join(self.base_url, f"release-{self.args.release}", "fasta", self.args.specie, item)
        file_name = self.get_link(url, item)
        link = os.path.join(url, file_name)

        ### Download fasta files
        fasta_path = self.wget_fasta(link, file_name)
        self.downloaded_fasta.append(fasta_path)

        ### check fasta file
        self.check_fasta(fasta_path)

        ### create a temp file, with alternate chromosome removed
        temp_fasta_path = self.filtered_fasta(fasta_path)
        self.temp_fasta_files.append(temp_fasta_path)


    def get_link(self, url, item):
        if item == 'cdna':
            pattern = 'cdna.all.fa.gz'
        elif item == 'ncrna':
            pattern = 'ncrna.fa.gz'
        try:
            response = requests.get(url, timeout=10)
        except requests.exceptions.ConnectionError:
            sys.exit("Error: unable to connect to Ensembl API.")
        if response.ok:
            response_text = response.text
        else:
            return response.raise_for_status()
        soup = BeautifulSoup(response_text, 'html.parser')
        href = [link.get('href') for link in soup.find_all('a') if pattern in link.get('href')][0]
        return href


    def wget_fasta(self, link, file_name):
        gz_file = file_name.split('.')
        gz_file = f"{'.'.join(gz_file[0:2])}.{self.args.release}.{'.'.join(gz_file[2:])}"
        fasta_path = os.path.join(self.tmpdir, gz_file)
        urlretrieve(link, fasta_path)
        return fasta_path


    def check_fasta(self, fasta_file):
        try :
            with gzip.open(fasta_file) as fh:
                first_line = fh.readline().rstrip().decode()
                if not first_line.startswith(">"):
                    sys.exit(f"{RED}Error: Are you sure {fasta} is a fasta file ?{ENDCOL}")
                return first_line
        except FileNotFoundError:
            sys.exit(f"{RED}Error: File '{fasta}' not found.{ENDCOL}")


    def filtered_fasta(self, fasta_file_path):
        ## Handle sequences
        sequences = []
        seq = []
        current_header = ""
        end_head =  '\n'            # '\t' if args.tsv else '\n'
        sep_seq  =  '\n'            # '' if args.uniq or args.tsv else '\n'
        with gzip.open(fasta_file_path, 'rt') as fh:
            previous_header = fh.readline()
            for line in fh:
                if line.rstrip():
                    if line.startswith('>'):
                        current_header = line
                        # ~ if not 'CHR_' in previous_header.split()[2]:
                        ### keep only transcripts in regular chromosomes
                        if previous_header.split()[2].split(':')[2] in self.chr:
                            ## add only the transcript name in the header (without version number)
                            sequences.append(f"{previous_header.split('.')[0]}\n{''.join(seq)}\n")
                        seq = []
                        previous_header = current_header
                    else:
                        seq.append(line.rstrip())
        ### last fasta sequence is not printed in the loop
        if previous_header.split()[2].split(':')[2] in self.chr:
        # ~ if not 'CHR_' in previous_header.split()[2]:
            sequences.append(f"{previous_header.split('.')[0]}\n{''.join(seq)}\n")
        ### write temp file
        fasta_file_basename = ".".join(os.path.basename(fasta_file_path).split('.')[:-2])
        temp_fasta_path = os.path.join(self.tmpdir, f'{fasta_file_basename}-altCHR.fa')
        with open(temp_fasta_path, 'w') as fh:
            for sequence in sequences:
                fh.write(sequence)
        ### check if temporary file done
        if not os.path.isfile(temp_fasta_path):
            sys.exit(f"Error: temporary {temp_fasta_path!r} not found.")
        return temp_fasta_path


    def mk_transcriptome(self):
        ### define output for filtered transcriptome
        specie, version, release = self.temp_fasta_files[0].split('.')[:3]
        specie = os.path.basename(specie).lower()
        transcriptome_fa = f"{specie}.{version}.{release}.transcriptome.fa"
        transcriptome_fa = os.path.join(self.args.datadir, transcriptome_fa)
        ### concatene cDNA and ncRNA
        if self.args.debug: print(f"{DEBUG}creating transcriptome {os.path.basename(transcriptome_fa)!r}.{ENDCOL}")
        with open(transcriptome_fa, 'w') as outfile:
            for fasta in self.temp_fasta_files:
                with open(fasta, 'r') as infile:
                    # ~ outfile.write(infile.read())        # small files
                    for line in infile:                     # big files
                        outfile.write(line)
        ''' Other method
        with open(transcriptome_fa, 'w') as fout, fileinput.input(temp_fasta_files, 'rb') as fin:
            for line in fin:
                fout.write(line)
        '''
        return transcriptome_fa


    def mk_index(self, transcriptome_fa, transcriptome_jf):
        """ Function doc """
        ### Select best tool to indexing transcriptome
        tools = ['_kmc', 'jellyfish']
        num = None
        for i,bin in enumerate(tools):
            if shutil.which(bin):
                num = i
                break
        if not num:
            print(f"Warning: no tool found to index the transcriptome")
            return None

        ### Define command line according to the tool used
        ''' in the future (python 3.10)
        match tools[num]:
            case 'kmc':
                print(f"prefered tool: {os.path.basename(tool)}")
            case 'jellyfish':
                cmd = f"{tool} count -m self.args.kmer_length -s 100000 {transcriptome_fa} -o {transcriptome_idx}"
        '''
        tool = tools[num]
        ## kmc case
        if tool == 'kmc': print(f"prefered tool: {os.path.basename(tool)}")
        ## jellyfish case
        elif tool == 'jellyfish':
            filename = os.path.basename(f"{os.path.splitext(transcriptome_fa)[0]}.jf")
            cmd = (f"{tool} count -t {self.args.thread} -m {self.args.kmer_length} -s 100000 "
                   f"{transcriptome_fa} -o {transcriptome_jf}")

        ### Build index
        if self.args.debug: print(f"{DEBUG}Build index with {tool!r}, please wait...{ENDCOL}")
        ret = os.system(cmd)


    def mk_pickle(self, transcriptome_fa):
        """ Build pickle file instead the fasta file (faster to load) """
        fa_dict = {}
        with open(transcriptome_fa) as fh:
            header = fh.readline()[1:].split()[0].rstrip()
            for line in fh:
                if line.startswith('>'):
                    fa_dict[header] = seq
                    header = line[1:].rstrip().split()[0].split('.')[0]
                else:
                    seq = line.rstrip()
        ### Don't forget the last record
        fa_dict[header] = seq
        ### how many records ?
        self.transcript_count = len(fa_dict)
        ### write as pickle file
        transcriptome_pkl = f"{os.path.splitext(transcriptome_fa)[0]}.pkl"
        with open(transcriptome_pkl, 'wb') as fh:
            pickle.dump(fa_dict, fh)


    def _to_report(self):
        """ Function doc """
        self.report.append("\n## Transcriptome Info\n")
        self.report.append(f"- nb transcripts: {self.transcript_count}\n")



### kmerize.py

"""
https://stackoverflow.com/questions/44587669/python-multiprocessing-how-to-close-the-multiprocessing-pool-on-exception
https://stackoverflow.com/questions/47903791/how-to-terminate-a-multiprocess-in-python-when-a-given-condition-is-met
"""

import os
import sys
import shutil
import subprocess
import multiprocessing
import copy
import random
import string

from color import *

MAX_CHARS = os.pathconf('/', 'PC_NAME_MAX') - 40   # with --fasta-file arg: avoid too large file name

class SpecificKmers:
    """ Class doc """

    def __init__(self, args, report, items, transcriptome_dict=None, geneinfo_dict=None):
        """
        1. Get sequence
        2. write on file
        3. From sequence, build abundance for each kmer, using jellyfish againt the genome (output: dict)
        4. From sequence, build abundance for each kmer, uding jellyfish againt the transcriptome (output: dict)
        5. filter the specific kmers, according to the given arguments
        6. write specific kmers and contigs in files
        """

        ### create a shared dict among multiple processes with Manager()
        manager = multiprocessing.Manager()
        self.args = manager.dict(args.__dict__)
        if args.selection:
            self.transcriptome_dict = manager.dict(transcriptome_dict)
            self.geneinfo_dict = manager.dict(geneinfo_dict)

        ### launch workers
        self.transcriptome_file = os.path.join(args.datadir, f"{args.specie}.{args.assembly}.{args.release}.k{args.kmer_length}.transcriptome.jf")
        try:
            with multiprocessing.Pool(processes=args.thread) as pool:
                if args.selection:
                    messages = pool.map(self.worker_selection, items)
                elif args.fasta_file:
                    messages = pool.map(self.worker_fasta_file, items)
            for type,mesg in messages:
                report[type].append(mesg)
        except KeyError as e:
            pool.close()
            shutil.rmtree(args.tmpdir)
            sys.exit(f"{RED}Error: {e.args[0]}")


    def worker_selection(self, item):
        """ Function doc """
        ### unpack item
        globals().update(item)
        global level
        level = 'transcript' if type == 'transcript' else 'gene'

        ### 1. Get sequence of the transcript (from the transcriptome)
        try:
            seq = self.transcriptome_dict[ENST]
        except KeyError:
            mesg = f"{given}: transcript not found in transcriptome ({ENST})."
            return 'failed', mesg
        if len(seq) < self.args['kmer_length']:
            mesg = f"{given}: sequence to short ({len(seq)} < {self.args['kmer_length']})."
            return 'failed', mesg

        ### 2. Write sequences in temporary file (needed by jellyfish)
        f_id = f"{item['given']}.{item['ENST']}"
        seq_file, fasta_name = self.dump_seq(f_id, seq)

        ### 3. From sequence, compute jellyfish againt the genome/transcriptome and convert results as dict ()
        kmercounts_genome_dict = self.jellyfish(seq_file, self.args['genome'])
        kmercounts_transcriptome_dict = self.jellyfish(seq_file, self.transcriptome_file)
        ### 4. filter the specific kmers, according to the arguments
        mesg = self.get_specific_kmers(item, kmercounts_transcriptome_dict, kmercounts_genome_dict, fasta_name)
        return mesg


    def worker_fasta_file(self, item):
        ### unpack item
        globals().update(item)
        ### 1. write separate file for each item
        seq_file, fasta_name = self.dump_seq(item['f_id'], item['seq'])

        ### 2. From sequence, compute jellyfish against the genome/transcriptome and convert results as dict ()
        kmercounts_transcriptome_dict = self.jellyfish(seq_file, self.transcriptome_file)
        kmercounts_genome_dict = self.jellyfish(seq_file, self.args['genome'])

        ### 3. find to specific kmers
        mesg = self.get_specific_kmers(item, kmercounts_transcriptome_dict, kmercounts_genome_dict, fasta_name)

        return mesg


    def dump_seq(self, f_id, seq):
        '''
        Write sequences in temporary file (needed by jellyfish)
        '''
        seq_dir = os.path.join(self.args['tmpdir'], 'sequences')
        os.makedirs(seq_dir, exist_ok=True)
        ### replace forbiden characters
        fasta_name = f_id.replace('/', '--')
        ### handle max file lenght
        if len(fasta_name) > MAX_CHARS:
            fasta_name = f"{f_id[:MAX_CHARS-15]}...{''.join(random.choices(string.ascii_letters, k=11))}"
        ### write sequence as fasta
        seq_file = os.path.join(seq_dir, f"{fasta_name}.fa")
        with open(seq_file, 'w') as fh:
            fh.write(f">{fasta_name}\n{seq}\n")
        return seq_file, fasta_name


    def jellyfish(self, seq_file, jf_file):
        """
        From sequence, compute jellyfish againt the genome/transcriptome and convert results
        as dict ()
        """
        if self.args['debug']: print(f"{YELLOW}start jellyfish on {os.path.basename(seq_file)} "
                                     f"against {os.path.basename(jf_file)}{ENDCOL}")
        cmd = f'jellyfish query -s "{seq_file}" {jf_file}'
        try:
            result = subprocess.check_output(cmd, shell=True, text=True, stderr=subprocess.STDOUT) \
                                             .rstrip('\n').split('\n')
        except subprocess.CalledProcessError as err:
            sys.exit(f"{ERROR}Error: executing jellyfish:\n"
                     f"  {ERROR}command: {ENDCOL}{cmd}\n"
                     f"  {ERROR}returned: {ENDCOL}{err.output}")
        if not result[0]:
            return {}
        result_dict = dict([(b[0], int(b[1])) for b in [a.split() for a in result]])
        return result_dict


    def get_specific_kmers(self, item, kmercounts_transcriptome_dict, kmercounts_genome_dict, fasta_name):
        '''
        Keep only specific kmers, according to the arguments
        Launched for each gene, transcript, or unanotated sequence
        '''
        args = copy.deepcopy(self.args)
        ### Define some variables: gene_name, transcript_name, variants_dic and output file names
        '''
        d = {
            'specific_kmers': [],              # specific kmers list
            'specific_contigs': [],            # specific contigs list
            masked_kmers = []                  # masked kmers list
            'contig': "",                      # initialize contig sequence
            'knb': 0,                          # kmer number (selected kmer)
            'c_nb': 1,                         # contig number
            'kmer_pos_prev': 0,                # position of retained kmer
            'contig_pos': 0,                   # position of retained contig
        }
        '''
        level = 'gene' if item['type'] != 'transcript' else item['type']
        specific_kmers = []              # specific kmers list
        masked_kmers = []                # masked kmers list
        specific_contigs = []            # specific contigs list
        contig = ""                      # initialize contig sequence
        knb = 0                          # kmer number (selected kmer)
        c_nb = 1                         # contig number
        kmer_pos_prev = 0                # position of retained kmer
        contig_pos = 0                   # position of retained contig

        ### initialization of count variables
        total_kmers = len(kmercounts_transcriptome_dict)
        if args['selection']:
            given_up = given.upper()
            isoforms = self.geneinfo_dict['gene'][ENSG]['transcript']
            isoforms_nb = len(isoforms)
            isoforms_dict = { isoform:self.transcriptome_dict[isoform] for isoform in isoforms }

            ### Define conditionnal variables
            ## When '--selection' option is set
            kmer_outfile = f"{given_up}-{ENST}-{level}-specific_kmers.fa"
            contig_outfile = f"{given_up}-{ENST}-{level}-specific_contigs.fa"
            masked_outfile = f"{given_up}-{ENST}-{level}-masked_kmers.fa"
        # ~ elif args['chimera']:
            # ~ ## When '--chimera' option is set
            # ~ kmer_outfile = f"{fasta_name}-chimera-specific_kmers.fa"
            # ~ contig_outfile = f"{fasta_name}-chimera-specific_contigs.fa"
        elif args['fasta_file']:
            ## When '--fasta-file' option is set
            kmer_outfile = f"{fasta_name}-transcript-specific_kmers.fa"
            contig_outfile = f"{fasta_name}-transcript-specific_contigs.fa"
            masked_outfile = f"{fasta_name}-masked_kmers.fa"


        i = 1
        for mer, abund_in_tr in kmercounts_transcriptome_dict.items():

            kmer_pos = i
            try:
                canonical = self.__canonical(mer)
                abund_in_ge = int(kmercounts_genome_dict[canonical])    # abundance in genome for this kmer
            except KeyError as err:
                if len(mer) != len(next(iter(kmercounts_genome_dict))):
                    raise KeyError(f"ErrorIndexLength: length of kmer expected: {self.args['kmer_length']}\n"
                                   f"  Genome kmer index length: {len(next(iter(kmercounts_genome_dict)))}\n"
                                   f"  Transcriptome kmer index length (dataset): {len(mer)}")
                raise KeyError(f"Error: kmer not found in genome: {err}")

            if level == 'gene':

                isoforms_with_mer = [enst for enst, seq in isoforms_dict.items() if mer in seq]
                isoforms_with_mer_nb = len(isoforms_with_mer)

                ### if the kmer is present/unique or does not exist (splice/chimera?) on the genome
                if abund_in_ge <= 1:

                    # knb, specific_kmers, given_up, transcript, kmer_pos, isoforms_with_mer_nb,
                    # isoforms_nb, mer, contig, kmer_pos_prev, kmer_pos, contig_pos

                    ### all kmers found (abund_in_tr) must be in isoforms
                    if not args['stringent'] and abund_in_tr and abund_in_tr == isoforms_with_mer_nb:
                        ## contigs case
                        if knb == 0:                                    # first kmer in contig
                            contig = mer
                            kmer_pos_prev = kmer_pos
                            contig_pos = kmer_pos
                        elif knb>0 and kmer_pos == kmer_pos_prev+1:     # add last base to existing contig
                            contig += mer[-1]
                            kmer_pos_prev = kmer_pos
                        else:                                           # store contig and create new
                            specific_contigs.append(f">{given_up}:{ENST}.contig_{c_nb} (at position {contig_pos})\n{contig}")
                            c_nb += 1
                            contig = mer
                            kmer_pos_prev = kmer_pos
                            contig_pos = kmer_pos
                        ## kmers case
                        knb += 1
                        specific_kmers.append(f">{given_up}:{ENST}.kmer{kmer_pos} ct:{c_nb} tr:{isoforms_with_mer_nb}/{isoforms_nb}\n{mer}")

                    ### for gene level only, the stringent argument implies retaining kmers present in ALL isoforms of the gene
                    elif args['stringent'] and abund_in_tr == isoforms_nb == isoforms_with_mer_nb:
                        ## contigs case
                        if knb == 0:                                    # first kmer in contig
                            contig = mer
                            kmer_pos_prev = kmer_pos
                            contig_pos = kmer_pos
                        elif knb>0 and kmer_pos == kmer_pos_prev+1:     # add last base to existing contig
                            contig += mer[-1]
                            kmer_pos_prev = kmer_pos
                        else:                                           # store contig and create new
                            specific_contigs.append(f">{given_up}:{transcript}.contig_{c_nb} (at position {kmer_pos})\n{contig}")
                            c_nb += 1
                            contig = mer
                            kmer_pos_prev = kmer_pos
                            contig_pos = kmer_pos
                        ## kmers case
                        knb += 1
                        specific_kmers.append(f">{given_up}:{ENST}.kmer{kmer_pos} ct:{c_nb} tr:{isoforms_with_mer_nb}/{isoforms_nb}\n{mer}")
                    else:
                        masked_kmers.append(f">{given_up}:{ENST}.kmer{kmer_pos} "
                                        f"tr:{isoforms_with_mer_nb}/{isoforms_nb} "
                                        f"transcriptome:{abund_in_tr}\n{mer}"
                                        )

                ### the kmer count exceeded 1 in the genome
                else:
                    masked_kmers.append(f">{given_up}:{ENST}.kmer{kmer_pos} "
                                    f"tr:{isoforms_with_mer_nb}/{isoforms_nb} "
                                    f"genome:{abund_in_ge}\n{mer}"
                                    )

            ### Cases of transcripts 1) annotated, 2) unannotated.
            elif level == 'transcript':
                ### Case of annotated transcripts
                if args['selection'] and abund_in_tr == 1 and abund_in_ge <= 1:
                    ## contigs case
                    if knb == 0:                                    # first kmer in contig
                        contig = mer
                        kmer_pos_prev = kmer_pos
                        contig_pos = kmer_pos
                    elif knb>0 and kmer_pos == kmer_pos_prev+1:     # add last bp to existing contig
                        contig += mer[-1]
                        kmer_pos_prev = kmer_pos
                    else:                                           # store contig and create new
                        specific_contigs.append(f">{ENST}.contig_{c_nb} (at position {contig_pos})\n{contig}")
                        c_nb += 1
                        contig = mer
                        kmer_pos_prev = kmer_pos
                        contig_pos = kmer_pos
                    ## kmers case
                    knb += 1
                    specific_kmers.append(f">{given_up}:{ENST}.kmer{kmer_pos} ct:{c_nb}\n{mer}")

                ### Case of unannotated transcripts
                elif args['fasta_file'] and abund_in_tr <= self.args['max_on_transcriptome'] and abund_in_ge <= self.args['max_on_genome']:     # max_on_transcriptome = 0 by default
                    ### contigs case
                    if knb == 0:                                    # first kmer in contig
                        contig = mer
                        kmer_pos_prev = kmer_pos
                        contig_pos = kmer_pos
                    elif knb>0 and kmer_pos == kmer_pos_prev+1:     # add last bp to existing contig
                        contig += mer[-1]
                        kmer_pos_prev = kmer_pos
                    else:
                        specific_contigs.append(f">{f_id}.contig_{c_nb} (at position {contig_pos})\n{contig}")
                        c_nb += 1
                        contig = mer
                        kmer_pos_prev = kmer_pos
                        contig_pos = kmer_pos
                    ### kmers case
                    knb += 1
                    specific_kmers.append(f">{f_id}.kmer{kmer_pos} ct:{c_nb}\n{mer}")

                ### the kmer count exceeded 1 in the genome
                else:
                    fasta_id = f"{given_up}:{ENST}.kmer{kmer_pos}" if args['selection'] else f"{f_id}.contig_{c_nb}"
                    masked_kmers.append(f">{fasta_id} "
                                    f"genome:{abund_in_ge} "
                                    f"transcriptome:{abund_in_tr}\n{mer}"
                                    )

            # ~ ### Case of chimera
            # ~ elif level == 'chimera':
                # ~ if abund_in_tr == abund_in_ge == 0:
                    # ~ ### contig case
                    # ~ if knb == 0:
                        # ~ contig = mer
                        # ~ kmer_pos_prev = kmer_pos
                        # ~ contig_pos = kmer_pos
                    # ~ elif knb>0 and kmer_pos == kmer_pos_prev+1:
                        # ~ contig += mer[-1]
                        # ~ kmer_pos_prev = kmer_pos
                    # ~ else:
                        # ~ specific_contigs.append(f">{f_id}.contig_{c_nb} (at position {contig_pos})\n{contig}")
                        # ~ c_nb += 1
                        # ~ contig = mer
                        # ~ kmer_pos_prev = kmer_pos
                        # ~ contig_pos = kmer_pos
                    # ~ ### kmers case
                    # ~ knb += 1
                    # ~ specific_kmers.append(f">{f_id}.kmer{kmer_pos} ct:{c_nb}\n{mer}")
            ### not a gene or transcript
            else:
                raise KeyError(f"{RED}Error: level {level!r} unknown.{ENDCOL}")
            i += 1


        ### append last contig in list
        if args['fasta_file'] and contig:
            specific_contigs.append(f">{f_id}.contig_{c_nb} (at position {contig_pos})\n{contig}")
        elif level == "gene" and contig:
            specific_contigs.append(f">{given_up}:{ENST}.contig_{c_nb} (at position {contig_pos})\n{contig}")
        elif level == "transcript" and contig:
            specific_contigs.append(f">{ENST}.contig_{c_nb} (at position {contig_pos})\n{contig}")
        if args['debug']:
            if args['selection']:
                print(f"{YELLOW} {ENST} kmers/contig: {len(specific_kmers)}/{len(specific_contigs)} ({given}){ENDCOL}")
            else:
                print(f"{YELLOW} {f_id} kmers/contig: {len(specific_kmers)}/{len(specific_contigs)}{ENDCOL}")

        ### masked kmers
        if masked_kmers:
            self.write(args, masked_outfile, masked_kmers, 'masked')

        ### write kmer/contig files
        if specific_kmers:
            self.write(args, kmer_outfile, specific_kmers, 'kmers')
            self.write(args, contig_outfile, specific_contigs, 'contigs')
        else:
            if args['selection']:
                mesg = f"{given}: no specific kmers found for {ENST} (level: {level})"
            else:
                mesg = f"{f_id}: no specific kmers found"
            return 'failed', mesg

        ### report
        if args['selection']:
            mesg = f"{given}: {item['symbol']}:{ENST} - kmers/contigs: {len(specific_kmers)}/{len(specific_contigs)} (level: {level})"
        else:
            mesg = f"{f_id} - kmers/contigs: {len(specific_kmers)}/{len(specific_contigs)} (level: {level})"

        return 'done', mesg


    def write(self, args, outfile, specific_seq, type):
        """ Write results as fasta file"""
        outdir = os.path.join(args['tmpdir'], type)
        os.makedirs(outdir, exist_ok=True)
        with open(os.path.join(outdir, outfile.replace('/', '--')), 'w') as fh:
            fh.write("\n".join(specific_seq) + '\n')


    def __canonical(self, mer):
        revcomp = mer[::-1].translate(str.maketrans('ATCG','TAGC'))
        return sorted((mer, revcomp))[0]


'''
def handle_kmers(**kwargs):
    print("LOCALS():", locals())
    ## kmers case
    knb += 1
    specific_kmers.append(f">{given_up}:{given_up}.kmer{kmer_pos} ({isoforms_with_mer_nb}/{isoforms_nb})\n{mer}")
    ## contigs case
    if knb == 1:                                    # first kmer in contig
        contig = mer
        kmer_pos_prev = kmer_pos
        print(f"{given_up}:{transcript}.contig{c_nb}\n{contig}")
    elif knb>1 and kmer_pos == kmer_pos_prev+1:     # add last pb to existing contig
        contig = contig + mer[-1]
        kmer_pos_prev = kmer_pos
    else:                                           # store contig and create new
        specific_contigs.append(f"{given_up}:{transcript}.contig{c_nb}\n{contig}")
        c_nb += 1
        contig = mer
        kmer_pos_prev = kmer_pos
'''




import os
import sys
import subprocess
from configparser import ConfigParser, ParsingError

import info
import color

DEFAULT_CONFIG = """
[CMD_ARGS]
## --datadir option
## path to kmerator datasets, one per release:
##   - modified transcriptome file (pkl)
##   - jellyfish transcriptome file (jf))
##   - metadata files (pkl)
##   - report file (md)
# datadir = /path/to/kmerator/directory

## --release option
##  release number (ex: 109) or last (last find for last release)
## default: last
# release = last

## --genome option
## Path of genome jellyfish index
# genome = /index/jellyfish/GRCh38_with_MT.jf

## --specie option
## set a specie (default: human)
# specie = human

## --kmer-length option
## set kmer length (default: 31)
# kmer_length = 31

## --max-on-transcriptome
## Only with --fasta-file option
# --max-on-transcriptome = 0

## --max-on-genome
## Only with --fasta-file option
# --max-on-genome = 1

## --stringent option
## Only with --selection option
# stringent = False

## output directory
# output = ./output

## --yes option
## assumes 'yes' as the prompt answer, run non-interactively
## default: False
# yes = True

## --thread option
## number of thread (default: 1)
# thread = 1

## --keep option
## keep intermediate files (default: False)
# keep = False
"""

class Config:
    """
    attributes:
        - filename: name of configuration file
        - filepath: path of the configuration file
        - content: dict of configuration file
    methods:
        - edit(): edit configuration file
    """
    filename = 'config-v3.ini'

    def __init__(self, appname):
        """ Class initialiser """
        ### define location of config file
        self.filePath = self.get_configPath(appname)
        ### create config file if not exists
        if not os.path.isfile(self.filePath):
            self._set_default()
        ### create a dict of config file
        self.content = ConfigParser()
        try:
            self.content.read(self.filePath)
        except  ParsingError as err:
            for line, text in err.errors:
                print(f"{color.RED}❗Syntax error at line {line} ({text})")
            ithem = 'it' if len(err.errors) == 1 else 'them'
            print(f"Please correct {ithem} in {err.source!r} file.")
            sys.exit()


    def get_configPath(self, appname):
        """
        Find to config
        """
        ### Define config file location (user or root)
        if os.geteuid():
            configdir = os.path.join(os.path.expanduser('~'), '.config', appname)
        else:
            configdir = os.path.join('/etc', appname)
        configfile = os.path.join(configdir, self.filename)
        return configfile


    def _set_default(self):
        try:
            os.makedirs(os.path.dirname(self.filePath), exist_ok=True)
            with open(self.filePath, 'w') as fs:
                fs.write(DEFAULT_CONFIG)
        except PermissionError as err:
            sys.exit(err.msg)


    def edit(self):
        """ Edit configuration file """
        ### define editor
        editor = 'editor'
        if os.environ.get('XDG_SESSION_TYPE') == 'x11':
            editor = 'xdg-open'
        elif 'EDITOR' in os.environ:
            editor = os.environ['EDITOR']
        ### launch config edit
        subprocess.call([editor, self.filePath])



#!/usr/bin/env python3

"""
Give some information about gene
examples:
  kmerator --info npm1 ensg00000268555 ENSG00000258154
  kmerator --info file.txt (list of items separated by space/tab/newline. comment: #)
"""

import sys
import os
import pickle

### Local modules
from dataset import Dataset
import color as col
import exit


def info(args):
    """ Function doc """

    ### load geneinfo pickle file as dict
    dataset = Dataset(args)
    geneinfo_dict = dataset.load_geneinfo()
    transcriptome_dict = dataset.load_transcriptome() if args.all else None
    ### get info for each item
    info, not_found = get_info(args, geneinfo_dict)
    ### print info
    output(args, geneinfo_dict, info, not_found, transcriptome_dict)
    ### quit removing temporary files
    exit.gracefully(args)


def get_info(args, geneinfo_dict):
    """
    according to the type of info (symbol, Ensembl gene, alias, Ensembl transcript),
    get information on the gene (or parent gene) from geneinfo_dict
    """
    info = {}
    not_found = []

    for given in args.info:
        ### if item given is a symbol
        if given.upper() in geneinfo_dict['symbol']:
            info[given] = {'_type': 'gene symbol'}
            ENSGs = geneinfo_dict['symbol'][given.upper()]
            for ensg in ENSGs:
                info[given][ensg] = geneinfo_dict['gene'][ensg]
        ### if item given is an alias
        elif given.upper() in geneinfo_dict['alias']:
            info[given] = {'_type': 'alias'}
            ENSGs = geneinfo_dict['alias'][given.upper()]
            for ensg in ENSGs:
                info[given][ensg] = geneinfo_dict['gene'][ensg]
        ### if item given is a ensembl gene name (ex: ENSG00000117)
        elif given.upper() in geneinfo_dict['gene']:
            info[given] = {'_type': 'ensembl gene name'}
            info[given][given] = geneinfo_dict['gene'][given.upper()]
        ### if item given is a ensembl transcript name (ex:ENST00000633522 )
        elif given.upper() in geneinfo_dict['transcript']:
            info[given] = {'_type': 'transcript'}
            ensg = geneinfo_dict['transcript'][given.upper()]
            info[given][given.upper()] = geneinfo_dict['gene'][ensg]
        else:
            not_found.append(given)

    return info, not_found


    for given in args.info:
        found_count = 0
        found_transcripts = []
        item = given.upper()

        if item in geneinfo_dict['transcript']:
            found_count += 1
            ENSG = geneinfo_dict['transcript'][item]
            items.append({'given':  given, 'ENST': item, 'type': 'transcript',
                                'ENSG':   ENSG,
                                'symbol': geneinfo_dict['gene'][ENSG]['symbol'],
                                })
        else:
            report['failed'].append(f"{given}: not found in transcriptome")

        if found_count > 1:
            report['multiple'].append(f"{given}: {found_count} ({', '.join([i for i in found_transcripts])})")


def output(args, geneinfo_dict, info, not_found, transcriptome_dict):
    """ Function doc """

    ### output as readable format
    for given, ENSGs in info.items():
        number_found = "" if ENSGs['_type'] == "ensembl gene name" else f"({len(ENSGs)-1} found) "

        print(f"\n=== {given} ({ENSGs['_type']}) {number_found} ===")
        given_type = ENSGs.pop('_type')     # gene symbol, alias, ENSG, transcript

        for ensg, val in ENSGs.items():
            blank = "\n" + 23 * " "
            print(f"  Ensembl ID           {ensg}")
            print(f"  Gene Name            {val.get('symbol', '')}")
            if args.all: print( "  Aliases              {}".format(blank.join(val.get('aliases', ''))))
            print(f"  Specie               {args.specie}")
            print(f"  Assembly             {geneinfo_dict['assembly']}")
            print(f"  Coordinates          {val['chr']}:{val['start']}-{val['end']}")
            print(f"  Strand               {val['strand']}")
            print(f"  Canonical transcript {val['canonical']}")
            print(f"  Biotype              {val.get('biotype', 'unknown')}")
            print(_text_format(23, '  Description', val.get('desc', 'unknown')))
            ### Print transcrpts (or not)
            if args.all:
                if given_type == "transcript":
                    transcript = next(iter(info))
                    seq = transcriptome_dict.get(transcript, f"{col.RED}✘✘✘{col.ENDCOL} (sequence not found)")
                    length = f" ({len(seq)})" if seq[0] in ['A', 'T', 'C', 'G'] else ""
                    print(_text_format(23, f"{transcript}{length}", seq, pos_key='top'))
                else:
                    print(f"  Transcripts of {val.get('symbol', ensg)}")
                    transcripts = val['transcript']
                    transcripts.remove(val['canonical'])    # I want canonical at first
                    transcripts.insert(0, val['canonical']) # I want canonical at first
                    for transcript in transcripts:
                        seq = transcriptome_dict.get(transcript, f"{col.RED}✘✘✘{col.ENDCOL} (sequence not found)")
                        length = f" ({len(seq)})" if seq[0] in ['A', 'T', 'C', 'G'] else ""
                        # ~ print(_text_format(23, f"{transcript}{length}", seq, pos_key='top'))
                        print(f"{transcript}{length}\n{seq}")
            else:
                print(_text_format(23, f"  Transcripts ({len(val['transcript'])})", ' '.join(val['transcript'])))
            print()

    ### output not found items
    if not_found:
        print("=== NOT FOUND ===\n", *[f" {i}\n" for i in not_found])


def _text_format(marg:int, key:str , text:str, pos_key='left'):
    """ Function doc """
    size = os.get_terminal_size(2)[0] - marg                        # text chunks size
    ltext = [text[i:i+size] for i in range(0, len(text), size)]     # chunks
    smarg = '\n' + marg * ' '                                       # delimiter for chunks

    if pos_key == 'top':
        return f"{smarg[1:]}{key}{smarg}{smarg.join(line.lstrip(' ') for line in ltext)}"
    else:
        return f"{key:<{marg}}{smarg.join(line.lstrip(' ') for line in ltext)}"



# info.py

"""
Genral informations on to the program.
"""

APPNAME = "kmerator"
VERSION = "2.3.2"
VERSION_DATASET = 1
SHORTDESC = "Find specific gene or transcript kmers. And more."
LICENCE = "GPL3"
AUTHOR = 'Sébastien RIQUIER, IRMB, Montpellier'
AUTHOR_EMAIL = "sebastien.riquier@ucd.ie"
CONTIBUTORS = [
    'Chloe BESSIERE chloe.bessiere@inserm.fr>'
    'Benoit GUIBERT <benoit.guibert@inserm.fr>',
]
DOC = f"""
-------------------------------------------------
                                   _
 ____  __.                        | |
|    |/ _| ___ _ _   __  ___  __ _| |_ ___  _ __
|      <  |  _` ` \ / _ \ '__/ _` | __/ _ \| '__|
|    |  \ | | | | ||  __/ |   (_| | || (_) | |
|____|__ \|_| |_| |_\___|_|  \__,_|\__\___/|_|
        \/
 Version: v{VERSION}
 Dependencies:
   - Jellyfish >= v2.0
-------------------------------------------------
"""

EXAMPLES = f"""
------------------------------------------------------
EXAMPLES:

Before all, remember that kmerator needs a jellyfish index of the genome.

Good idea before requests kmerator:
 kmerator -e                     # Edit config file to set default options

Some requests:
 kmerator -s npm1 braf           # get specific kmers from NPM1 and BRAF genes
 kmerator -s genes.txt           # you can alse use a file with gene list (#: comment)
 kmerator -f file.fa             # give a fasta file from unannotated sequences

Maintains yours kmerator indexes
 kmerator -l                     # list local avalaible indexes
 kmerator --mk-dataset -r 101    # install dataset for release 101
 kmerator -u -S zebrafish        # update dataset if new release avalaible

Get info on some genes/transcript
 kmerator --info MLLT3 Braf      # info about the MLLT3 and BRAF genes
 kmerator --info MMLT3 --all     # extended info about the MMLT3 gene
 kmerator --info genes.txt       # genes/transcripts are in a file
"""



#!/usr/bin/env python3

'''
- if 'last' is given as release (default), get last release of transcriptome and it set to args.release
- check if dataset is present for this release (3 files)
- if not present, ask to build it
  - download some files from Ensembl database (mysql) and build 'geneinfo.pkl' file
  - download transcriptome (cdna + ncran), filter it and build a jellyfish index
  - write transcriptome as pickle file

--------
Ensembl links
  - http://www.ensembl.org/info/docs/api/core/core_schema.html
  - https://ftp.ensembl.org/pub/current-mysql/
  - http://www.ensembl.org/info/docs/api/core/diagrams/Core.svg
Other links:
  - https://lists.ensembl.org/pipermail/dev_ensembl.org/2013-January/003357.html
'''


import sys
import os
import argparse
import requests
from bs4 import BeautifulSoup
import gzip
import pickle
import threading
import subprocess

import info
from color import *
from mk_geneinfo import GeneInfoBuilder
from mk_transcriptome import TranscriptomeBuilder
import exit


species = {
    "human": "homo_sapiens",
    "mouse": "mus_musculus",
    'zebrafish': 'danio_rerio',
    "horse": "equus_caballus",
    "hen" : "gallus_gallus",
    "c.elegans": "caenorhabditis_elegans",
}


def main():
    """ Function doc """
    args = usage()
    dataset = Dataset(args)
    if args.list_dataset:
        dataset.list()
    elif args.update_last:
        dataset.update_last()
    elif args.rm_dataset:
        dataset.remove()
    elif args.mk_dataset:
        dataset.make()
    elif args.last_avail:
        dataset.last_available()
    elif args.load:
        geneinfo_dict = dataset.load_geneinfo()
        transcriptome_dict = dataset.load_transcriptome()
        for type in ('gene', 'symbol', 'alias', 'transcript'):
            print(type, next(iter(geneinfo_dict[type].items())))
        print(next(iter(transcriptome_dict.items())))
        print("Assembly in geneinfo:", geneinfo_dict['assembly'])
        print("Assembly in args.assembly:", args.assembly)
        print("Relesase", args.release)


class Dataset:
    """
    Methods:
      - __init__(): if release == 'last', redefine release with last release number
      - is_present(): Ensure release files are presents
      - build(): build files for the release
      - load_geneinfo(): load geneinfo file as dict
    """

    def __init__(self, args):
        """ Class initialiser """
        self.args = args
        self.base_url = "http://ftp.ensembl.org/pub"
        self.attended = ['transcriptome.pkl', 'transcriptome.jf', 'geneinfo.pkl', 'report.md']
        # ~ self.assembly = None
        self.transcriptome_fa = None         # transcriptome fasta path
        self.transcriptome_pkl = None        # transcriptome pickle path
        self.transcriptome_jf = None         # transcriptome jellyfish path
        self.geneinfo_pkl = None             # geneinfo path
        self.report_md = None                # report path
        self.report = []                     # report
        self.ebl_releases = []               # all releases avalaible on Ensembl
        self.dataset = self.set_dataset_dict()   # {complete: {specie: {release:[k31,...]}},partial:..., other:...}
        ### Is args.specie an alternative name ?
        if self.args.specie.lower() in species:
            self.args.specie =  species[self.args.specie.lower()]
        if self.args.release == 'last' and not self.args.list_dataset:
            self.args.release = self.get_ebl_current_release()
            self.get_ebl_releases()
        ### check if dataset is locally present and assign variable for each file
        if args.datadir and not self.args.list_dataset:
            self.dataset_ok = self.dataset_here()

    
    def set_dataset_dict(self):
        ### general variables
        files = next(os.walk(self.args.datadir))[2]
        attended = ['transcriptome.pkl', 'transcriptome.jf', 'geneinfo.pkl', 'report.md']
        dataset = {"complete": {},  # set complete 
                   "partial": {},   # set uncomplete, miss some files
                   "other": [],}     # files not part of dataset
        releases = {}
        
        ### dict of releases found: {<specie>: {<release>:[<file1>, <file2>]} }
        for file in files:
            try:
                file_l = file.split('.')
                specie = ".".join(file_l[:2])
                release = int(file_l[2])
                releases.setdefault(specie, {}).setdefault(release, []).append(file)
            except (IndexError, ValueError):
                dataset["other"].append(file)

        ### classify the releases (complete or incomplete)
        for specie in releases:
            for release, file_list in releases[specie].items():
                founded_items = set(".".join(f.split(".")[-2:]) for f in file_list)
                founded_files = list(map(lambda v: v in founded_items, self.attended))
                if all(founded_files):
                    dataset['complete'].setdefault(specie, {})[int(release)] = []
                else:
                    dataset['partial'].setdefault(specie, []).append(int(release))

        ### find for different version of jellyfish indexes of transcriptome
        for specie, rels in dataset['complete'].items():
            for rel in rels:
                for file in releases[specie][rel]:
                    if file.endswith('.transcriptome.jf'):
                        k = file.split('.')[3] if file.split('.')[3].startswith('k') else 'k?'
                        dataset['complete'][specie][rel].append(k)
        
        return dataset


    def get_ebl_current_release(self):
        """get number of last available release on Ensembl"""
        url = os.path.join(self.base_url, "current_mysql")
        try:
            r = requests.get(url)
        except requests.exceptions.ConnectionError:
            print(f"{YELLOW} Error connecting to Ensembl.{ENDCOL}")
            return None
        if not r.ok:
            print(r)
            return None
        soup = BeautifulSoup(r.text, 'lxml')
        core = [ a.text for a in soup.findAll('a') if a.text.startswith(f"{self.args.specie}_core")]
        ebl_current_release = core[0].split('_')[-2]
        return ebl_current_release


    def get_ebl_releases(self):
        """ get avalaible releases on Ensembl, limited to 90 """
        try:
            r = requests.get(self.base_url)
        except requests.exceptions.ConnectionError:
            ### If no connection to Ensembl, suggest to use last local Release
            print(f"{YELLOW} Error connecting to Ensembl.{ENDCOL}")
            local_releases = self.get_local_releases()
            if not local_releases:
                print(f"{YELLOW} No dataset found in {self.args.datadir!r}, exit.")
                exit.gracefully(self.args)
            ask = 'y'
            if not self.args.yes:
                ask = input(f" {YELLOW}Last local Release found in datasets is {max(local_releases)}. Continue with it? (Yn): {ENDCOL}") or 'y'
            if ask.lower() == 'y':
                self.ebl_releases = local_releases  # this is a lie :)
                return
            else:
                print("Aborted by user.")
                exit.gracefully(self.args)

        ### If an error occur connecting to Ensembl, use the last local release avalaible.
        if not r.ok:
            all_files = next(os.walk(self.args.datadir))[2]
            self.ebl_releases = {int(i.split('.')[2]) for i in all_files if i.split('.')[0] == self.args.specie}
            ask = 'y'
            if not self.args.yes:
                print(f"{YELLOW} Error connecting to Ensembl (code {r.status_code}).")
                if not self.ebl_releases:
                    print(f" No dataset found in {self.args.datadir!r}, exit.{ENDCOL}")
                    exit.gracefully(self.args)
                ask = input(f" {YELLOW}Do you want to continue with release {str(max(self.ebl_releases))!r}? (Yn): {ENDCOL}") or 'y'
            if ask.lower() == 'y':
                return
            else:
                print("Aborted by user.")
                exit.gracefully(self.args)

        soup = BeautifulSoup(r.text, 'lxml')
        self.ebl_releases = [int(a.text.split('-')[1].rstrip('/')) for a in soup.findAll('a') if a.text.startswith('release-')]
        self.ebl_releases = [n for n in self.ebl_releases if n >= 90]


    def get_local_releases(self):
        print("get_local_releases()", {int(i.split('.')[2]) for i in next(os.walk(self.args.datadir))[2] if i.split('.')[0] == self.args.specie} )
        return {int(i.split('.')[2]) for i in next(os.walk(self.args.datadir))[2] if i.split('.')[0] == self.args.specie}


    def load_geneinfo(self):
        """ Load geneinfo.pkl as dict"""
        ### if dataset not found, ask to install
        if not self.dataset_ok:
            self.build()
            self.dataset_ok = self.dataset_here()
        ### Load geneinfo
        with open(self.geneinfo_pkl, 'rb') as fic:
            geneinfo_dict = pickle.load(fic)
            self.args.assembly = geneinfo_dict['assembly']
        ### check version of dataset
        version_geneinfo = geneinfo_dict.get('version', 0)
        if version_geneinfo < info.VERSION_DATASET:
            print(f"{RED}\n ErrorVersion: application and dataset versions are not compatible.\n"
                   "               Please, reinstall the dataset:\n"
                  f"{YELLOW} {info.APPNAME} --mk-dataset -S {self.args.specie} -r {self.args.release}\n")
            exit.gracefully(self.args)
        return geneinfo_dict


    def load_transcriptome(self):
        """ Load transcriptome.pkl as dict"""
        ### if dataset not found, ask to install
        if not self.dataset_ok:
            self.build()                                # build dataset
            self.dataset = self.set_dataset_dict()      # re-compute datasets
            self.dataset_ok = self.dataset_here()
        ### Load transcriptome
        with open(self.transcriptome_pkl, 'rb') as fic:
            transcriptome_dict = pickle.load(fic)
        return transcriptome_dict


    def dataset_here(self):
        """
        - Check if dataset is present in datadir
        - assign files names to matching variables
        - jump to rename_jf() if the releases jellyfish file names are in a older version (without kmer size)
         """
        def is_release_found(self):
            ### check if the specified release is present in the datasets
            releases_found = []
            for specie, releases in self.dataset['complete'].items():
                if specie.startswith(self.args.specie):
                    assembly = specie.split('.')[1]
                    releases_found = releases.get(int(self.args.release), [])
            if f"k{self.args.kmer_length}" in releases_found: 
                return releases_found, assembly, True
            else:
                return releases_found, assembly, False        
        releases_found, assembly, found = is_release_found(self)
        
        ### Handle older file format (version < 2.3.0)
        if 'k?' in releases_found:
            rename = input(
                        f"{YELLOW}The dataset has an unspecified kmer value in the jellyfish file "
                        "name, because it was created with an older version of kmerator. Kmerator "
                        f"can rename the files to the new format before continuing [Yn]:{ENDCOL} "
                        ).lower() or 'y'
            if rename == 'y':
                self.rename_jf()
                releases_found, assembly, found = is_release_found(self)
                
        ### Assign name to file attributes
        if found:
            basename = f"{self.args.specie}.{assembly}.{self.args.release}"
            pathbasename = os.path.join(self.args.datadir, basename)
            self.transcriptome_fa = f"{pathbasename}.transcriptome.fa"
            self.transcriptome_pkl = f"{pathbasename}.transcriptome.pkl" 
            self.transcriptome_jf = f"{pathbasename}.k{self.args.kmer_length}.transcriptome.jf"
            self.geneinfo_pkl = f"{pathbasename}.geneinfo.pkl"
            self.report_md = f"{pathbasename}.report.md"
            
        return found


    def define_dataset(self):
        """
        Define the names of the local files for the specified dataset
        """

        ### Request Ensembl to get info (assembly name, release number, list of releases)
        if not self.ebl_releases:
            self.get_ebl_releases()
        if not int(self.args.release) in self.ebl_releases or not self.args.release.isnumeric():
            print(f"{ERROR} Error: {self.args.release!r} is not a valid Release (valid releases range from 90 to {max(self.ebl_releases)}).")
            exit.gracefully(self.args)

        ### Get assembly name
        url = f"{self.base_url}/release-{self.args.release}/fasta/{self.args.specie}/cdna/"
        r = requests.get(url)
        if not r.ok:
            print(f"{ERROR} Error: not a valid url: {url!r}.\n check for release ({self.args.release!r}) and specie ({self.args.specie!r}).{ENDCOL}")
            exit.gracefully(self.args)
        soup = BeautifulSoup(r.text, 'lxml')
        self.args.assembly = [a.text.split('.')[1] for a in soup.findAll('a') if a.text.startswith(self.args.specie.capitalize())][0]

        ### Assign files names of dataset
        basename = f"{self.args.specie}.{self.args.assembly}.{self.args.release}"
        self.geneinfo_pkl = os.path.join(self.args.datadir, f"{basename}.geneinfo.pkl")
        self.transcriptome_pkl = os.path.join(self.args.datadir, f"{basename}.transcriptome.pkl")
        self.transcriptome_jf = os.path.join(self.args.datadir, f"{basename}.k{self.args.kmer_length}.transcriptome.jf")
        self.report_md = os.path.join(self.args.datadir, f"{basename}.report.md")


    def build(self):
        """
        if dataset of this release is not present in the local directory of kmerator (datadir),
        we must build him, downloading some files and rearange them.
        """
        ### Ask user to download files
        valid = 'y' if self.args.yes else input(f"Dataset for release {self.args.release} ({self.args.specie}) not found, install it? (Yn) ")
        if valid.lower() in ['n', 'no']:
            print("Exited by user.")
            exit.gracefully(self.args)

        ### Check target directory permissions
        if not os.access(self.args.datadir, os.W_OK):
            print(f"{RED}\n Error: write acces denied to datadir ({self.args.datadir}).{ENDCOL}")
            exit.gracefully(self.args)

        ### Define target files
        if not self.dataset_ok:
            self.define_dataset()

        ### build kmerator dataset for the specie/release specified (multithreaded)
        geneinfo = GeneInfoBuilder(self.args, self.base_url, self.geneinfo_pkl, self.report)
        geneinfo.get_meta()         # first step : get meta info
        chr = geneinfo.data['chr']

        # ~ geneinfo_dict = geneinfo.build()
        # ~ TranscriptomeBuilder(self.args, self.base_url, self.transcriptome_jf, self.report, chr)

        ### build kmerator dataset for the specie/release specified (multithreaded)
        th1 = threading.Thread(target=TranscriptomeBuilder,
                               args=(self.args, self.base_url, self.transcriptome_jf, self.report, chr))
        th2 = threading.Thread(target=geneinfo.build)
        th1.start()
        th2.start()
        th1.join()
        th2.join()
        

        ### write report
        with open(self.report_md, 'w') as fh:
            fh.write(f"# Kmerator files for {self.args.specie}, release {self.args.release}\n")
            fh.write('\n'.join(self.report))


    def list(self):
        """ List local releases """
        ### Show releases
        print(f"\n {YELLOW}Location of datasets:{ENDCOL} {self.args.datadir}")
        if self.dataset['complete']:
            to_print = f"\n {YELLOW}Datasets found:{ENDCOL}\n"
            for specie, releases in self.dataset['complete'].items():
                rels = sorted([(int(k),v) for k,v in releases.items()])
                for rel in rels:
                    to_print += f"  - {specie.split('.')[0]}: {rel[0]} ({', '.join(sorted(rel[1]))})\n"
            print(to_print)
        else:
            print(f"\n {YELLOW}No releases found{ENDCOL}\n")
        if self.dataset['partial']:
            print(f" {YELLOW}Incompletes datasets:{ENDCOL}")
            for specie, releases in self.dataset['partial'].items():
                print(f"  - {specie.split('.')[0]}: {', '.join([str(i) for i in releases])}{ENDCOL}")    
        print()

        ### Other file found in dataset location
        if self.dataset['other']:
            print(f" {YELLOW}Files not part of a dataset:{ENDCOL}")
            print(*[f"  - {i}" for i in self.dataset['other']], sep="\n")
        print()
        
        ### Find for dataset built by old version of kmerator, and propose to update the datasets
        found = False
        rename = 'n'
        for species, releases in self.dataset['complete'].items():
            if found: break
            for release, values in releases.items():
                if 'k?' in values:
                    found = True
                    rename = input(f"{YELLOW}\nSome releases has a kmer size not determined (k?) "
                            "due to previous version of kmerator. Do you want to automatically "
                            f"correct their size? [Yn]: {ENDCOL}").lower().strip() or 'y'
                    break
        if rename == 'y':
            self.rename_jf()

        ### exit
        exit.gracefully(self.args)


    def remove(self):
        """ remove a release """
        ### list dataset files this specie/release
        release_files = []
        jellyfish_files = []
        for file in next(os.walk(self.args.datadir))[2]:
            l_file = file.split('.')
            try:
                specie, assembly, release = l_file[:3]
                if self.args.specie == specie and self.args.release == release:
                    release_files.append(file)
                    if l_file[-1] == 'jf' and l_file[-2] == 'transcriptome':
                        jellyfish_files.append(file)
            except ValueError:
                continue
        ### if release empty
        if not release_files:
            print(f"Dataset not found for {self.args.specie!r}, release {self.args.release!r}.")
            exit.gracefully(self.args)
        ### if multiple jellyfish for the release, remove only the transcriptome file
        if len(jellyfish_files) > 1:
            for file in jellyfish_files:
                k = file.split('.')[3]
                if k.startswith('k'):
                    if int(k[1:]) == self.args.kmer_length:
                        release_files = [file]
                        break
                elif k == 'transcriptome':
                    release_files = [file]

        ### if not files to delete
        if not release_files:
            print(f"Dataset not found for {self.args.specie!r}, release {self.args.release!r}.")
            exit.gracefully(self.args)
        ### Ask to remove
        resp = 'y'
        if not self.args.yes:
            print(f"\nspecie: {self.args.specie} - release: {self.args.release}\n")
            # ~ print("Files to delete", *to_delete, sep='\n - ')
            print("Files to delete", *release_files, sep='\n - ')
            resp = input("\nDelete files (Yn): ") or 'y'
        ### Remove
        if not resp.lower() == 'n':
            for file in release_files:
                os.remove(os.path.join(self.args.datadir, file))
        else:
            print("\nAborted by user.  ")
        exit.gracefully(self.args)


    def make(self):
        """ Download files and make dataset """
        resp = 'y'
        if self.dataset_ok and not self.args.yes:
            resp = input(f"Release {self.args.release} already exists, erase? (Yn): ") or 'y'
        if resp.lower() == 'y':
            # ~ self.args.release = str(release)           # because build() using self.args.release
            self.args.yes = True
            print(f" 🧬 Build kmerator dataset for {self.args.specie}, release {self.args.release}, please wait...")
            self.build()

        exit.gracefully(self.args)


    def update_last(self):
        """ Function doc """
        if not self.ebl_releases:
            self.get_ebl_releases()
        self.args.release = self.get_ebl_current_release()

        if not self.dataset_here():
            ask = 'y'
            if not self.args.yes:
                ask = input(f"Dataset for {self.args.specie} will be updated to release {self.args.release}, continue ? (Yn): ") or 'y'
            if ask.lower() == 'y':
                ### Build Dataset
                self.make()
            else:
                sys.exit("Aborted by user.")
        else:
            print(f"The last release for {self.args.specie} is {self.args.release}, nothing to do.")
        exit.gracefully(self.args)


    def last_available(self):
        print(self.get_ebl_current_release())
        exit.gracefully(self.args)


    def rename_jf(self):
        """
        Migrate dataset to new version: 
        Before version 2.3.0, the jellyfish index file name did not containt the size of the kmers.
        rename_jf aims to rename file, adding kmer size in the name, according the 2.3.0 version
        """
        files = next(os.walk(self.args.datadir))[2]
        ### get k value
        get_k = lambda res,i: res[i + 1] 
        # ~ print(files)
        for file in files:
            if file.endswith(".transcriptome.jf"):
                file_l = file.split(".")
                k_expected = file_l[3]
                if len(file_l) == 5 and k_expected == "transcriptome":
                    ### jellyfish keep the command line used to build the jf file.
                    cmd_get = f"jellyfish info {os.path.join(self.args.datadir, file)} | head -1"
                    res = subprocess.check_output(cmd_get, shell=True, text=True).split(' ')
                    for opt in ('-m', '--mer-len'):
                        k = f"k{get_k(res, res.index(opt))}"
                        break
                    file_l.insert(3, k)
                    new_name = '.'.join(file_l)
                    cmd_set = f"mv {os.path.join(self.args.datadir, file)} {os.path.join(self.args.datadir, new_name)}"
                    try:
                        subprocess.check_output(cmd_set, shell=True)
                    except Exception as e:
                        sys.exit(e)
        self.dataset = self.set_dataset_dict()          


def usage():
    parser = argparse.ArgumentParser()
    exclusive = parser.add_mutually_exclusive_group(required=True)
    ### OPTION
    parser.add_argument('-S', '--specie',
                        help=(
                            "indicate a specie referenced in Ensembl, to help, follow the link "
                            "https://rest.ensembl.org/documentation/info/species. You can use "
                            "the 'name', the 'display_name' or any 'aliases'. For example human, "
                            "homo_sapiens or homsap are valid (default: human)."
                            ),
                        default='human',
                        )
    parser.add_argument('-r', '--release',
                        help="release of transcriptome (default: last).",
                        default="last",
                        )
    exclusive.add_argument('-d', '--datadir',
                        help=(
                            "Directory where kmerator file are stored. Files are:"
                            "\n - fasta file of modified transcriptome (fa)"
                            "\n - jellyfish of this transcriptome (jf)"
                            "\n - metadata file like gene-symbols or aliases, "
                            ),
                        # ~ required = True,
                        )
    parser.add_argument('-l', '--list-dataset', '--datasets',
                        action="store_true",
                        help="list local releases",
                       )
    parser.add_argument('--debug',
                        action="store_true",
                        help="Show more",
                       )
    parser.add_argument('--load',
                        action="store_true",
                        help="load dataset",
                       )
    parser.add_argument('--tmpdir',
                        help="temporary dir",
                        default="/tmp",
                       )
    parser.add_argument('--rm-dataset',
                        action="store_true",
                        help="remove a dataset, according with --specie and --release options",
                       )
    parser.add_argument('--mk-dataset',
                        action="store_true",
                        help="make a dataset, according with --specie and --release options",
                       )
    parser.add_argument('-t', '--thread',
                        type=int,
                        help="thread number (default: 1)",
                        default=1,
                       )
    parser.add_argument('-u', '--update-last',
                        action="store_true",
                        help="builds a new dataset if a new version is found on Ensembl",
                       )
    parser.add_argument('-k', '--kmer-length',
                        type=int,
                        help="kmer length (default: 31)",
                        default=31,
                       )
    exclusive.add_argument('--last-avail', '--last-available',
                        action='store_true',
                        help="last release available on Ensembl",
                       )
    parser.add_argument('--keep',
                        action='store_true',
                        help=("keep kmerator transcriptome as fasta format."
                            ),
                        )
    parser.add_argument('-y', '--yes',
                        action='store_true',
                        help=("assumes 'yes' as the prompt answer, run non-interactively."),
                        )
    return parser.parse_args()


if __name__ == "__main__":
    main()



### Not used
### Not finished

def longest_transcript(args, gene_name, transcriptome_dict):
    if args.verbose: print(f"{'-'*9}\n{Color.YELLOW}Finding the longest variant for the gene {gene_name}.{Color.END}")
    transcripts_dict = { k:len(v) for (k,v) in transcriptome_dict.items() if k.startswith(f"{gene_name}:")}
    # ~ print(*[k for k in variants_dict], sep='\n')
    nb_transcripts = len(transcript_dict)
    if args.verbose: print(f"{Color.YELLOW}Number of variants: {nb_transcripts}")
    longest_transcript = None
    length = 0
    for k,v in transcript_dict.items():
        if v > length:
            length = v
            longest_transcript = ':'.join(k.split(':')[1:2])
    # ~ print(f"{longest_variant = }")
    return longest_transcript



#!/usr/bin/env python3

import setuptools
from kmerator import info

setuptools.setup(
    name = 'kmerator',
    version = info.VERSION,
    author = info.AUTHOR,
    author_email = info.AUTHOR_EMAIL,
    description = info.SHORTDESC,
    long_description = open('README.md').read(),
    long_description_content_type = "text/markdown",
    url="https://github.com/Transipedia/kmerator",
    packages = setuptools.find_packages(),
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    entry_points = {
        'console_scripts': [
            'kmerator = kmerator.kmerator:main',
        ],
    },
    include_package_data = True,
    install_requires=['bs4', 'lxml', 'requests'],
    python_requires = ">=3.7",
    licence = "GPLv3"
)



#### exit.py
import sys
import shutil
from color import *


def gracefully(args):
    ### quit properly
    if not args.keep:
        try:
            shutil.rmtree(args.tmpdir, ignore_errors=True)
        except:
            exit(f"Error deleting {args.tmpdir}.{ENDCOL}")
    sys.exit(0)



import os
import requests
from bs4 import BeautifulSoup
import gzip
import pickle

import info
from color import *


class GeneInfoBuilder:
    """
    get files on Ensembl (mysql):
      - gene.txt
      - transcript.txt
      - xref.txt  (= gene-symbols)
      - external-synonym.txt (= aliases)
    merge info from theses files, and write has pickle file
    -----
    Example of result whith HSTF1 (alias):

    assembly: GRCh38
    ensg:
        ENSG00000185122:
            gene-symbol: HSF1
            canonical: ENST00000528838
            nb_transcript: 14
            chr: 11
            start: 144291591
            end: 144314720
            strand: 1
            aliases:
                - HSTF1
            enst:
                id: ENST00000528838
                seq: ATCG...
            enst:
                id:ENST00000400780
                seq: GCTA...
            ...
    transcript:
        ENST00000528838: ENSG00000185122
        ENST00000400780: ENSG00000185122
    symbols:
        HSF1: ENSG00000185122
        FGF4: ENSG00000075388
    aliases:
        HSTF1:
            - ENSG00000185122
            - ENSG00000075388
    -----
    http://www.ensembl.org/info/docs/api/core/core_schema.html#gene
    https://ftp.ensembl.org/pub/release-106/mysql/homo_sapiens_core_106_38/
    (http://www.ensembl.org/info/docs/api/core/diagrams/Core.svg)
    """

    def __init__(self, args, url, outfile, report):
        """ Class initialiser """
        if args.debug: print(f"{DEBUG}Build geneinfo, please wait...{ENDCOL}")
        self.args = args
        self.outfile = outfile
        self.report = report
        ### internal variables
        self.chr_dict = {}                      # ID:name for chromosomes
        ### get base url of files to download
        url = f"{url}/release-{args.release}/mysql/"
        r = requests.get(url)
        soup = BeautifulSoup(r.text, 'lxml')
        dir = [a.text for a in soup.findAll('a') if a.text.startswith(f"{args.specie}_core_{args.release}")][-1]
        self.url = url + dir
        ### dicts (refs: content downloaded, data: dict of results)
        self.refs = {"gene": {}, "xref": {}, "synonym": {}, "transcript": {}}
        self.data = {"gene": {}, "symbol": {}, "alias": {}, "transcript": {}, "assembly": "",
                     "chr": [], "version": info.VERSION_DATASET}


    def get_meta(self):
        '''
        2. get seq_region.txt to make a list of region_id (chromosomes + MT)
        1. get Assembly version of the specie
        AIMS - filter alternates chromosomes (scaffold, CHR_)
          - $0 = region_id
          - $1 = chromosome (ex: 16, X)
        push chromosomes list in geneinfo_dict['chr']
        push Assembly version in geneinfo_dic['assembly']

        https://lists.ensembl.org/pipermail/dev_ensembl.org/2013-January/003357.html
        '''
        if __name__ == "__main__": print("Get meta info, please wait...")

        '''
        CHROMOSOME LIST & ASSEMBLY
        https://lists.ensembl.org/pipermail/dev_ensembl.org/2013-January/003357.html
        '''
        ### get attrib_type file and keep id when second field == "karyotype_rank"
        attrib_type_url = f"{self.url}attrib_type.txt.gz"
        attrib_type_r = requests.get(attrib_type_url)
        attrib_type_str = gzip.decompress(attrib_type_r.content).decode()
        attrib_type_id = None
        for line in attrib_type_str.splitlines():
            line = line.split()
            if line[1] == "karyotype_rank":
                attrib_type_id = line[0]
                break
        del attrib_type_r, attrib_type_str

        ### get seq_region_attrib file and keep list of id when second field == attrib_type_id
        seq_region_ids = []
        seq_region_attrib_url = f"{self.url}seq_region_attrib.txt.gz"
        seq_region_attrib_r = requests.get(seq_region_attrib_url)
        seq_region_attrib_str = gzip.decompress(seq_region_attrib_r.content).decode()
        for line in seq_region_attrib_str.splitlines():
            line = line.split()
            if line[1] == attrib_type_id:
                seq_region_ids.append(line[0])
        del seq_region_attrib_r, seq_region_attrib_str

        ### get seq_region file and keep list of id when second field == attrib_type_id
        seq_region_url = f"{self.url}seq_region.txt.gz"
        seq_region_r = requests.get(seq_region_url)
        seq_region_str = gzip.decompress(seq_region_r.content).decode()
        for line in seq_region_str.splitlines():
            line = line.split()
            if line[0] in seq_region_ids:
                self.chr_dict[line[0]] = line[1]
        del seq_region_r, seq_region_str
        ## Add chromosome list in geneinfo
        self.data['chr'] = list(self.chr_dict.values())

        ### get meta file and keep assembly.default
        meta_url = f"{self.url}meta.txt.gz"
        meta_r = requests.get(meta_url)
        meta_str = gzip.decompress(meta_r.content).decode()
        for line in meta_str.splitlines():
            line = line.split()
            if line[2] == "assembly.default":
                assembly = line[3].split('.')[0]
                self.data["assembly"] = assembly
                break
        del meta_r, meta_str


    def build(self):
        """ Function doc """
        ### Get ENSG
        self._set_refs()
        ### Add SYMBOLS
        self._add_symbols()
        ### Add ALIASES
        self._add_aliases()
        ### Add TRANSCRIPTS
        self._add_transcripts()
        ### Write as file
        self._write()
        ### Add to report
        self._to_report()

        if self.args.debug:
            print(DEBUG)
            print("url:", self.url)
            print("Gene Info")
            print(f" nb genes: {len(self.data['gene'])}")
            print(f" nb symbols: {len(self.data['symbol'])}")
            print(f" nb aliases: {len(self.data['alias'])}")
            print(f" nb transcripts: {len(self.data['transcript'])}")
            print(ENDCOL)


    def _set_refs(self):
        if __name__ == "__main__": print("Get gene.txt, please wait...")
        '''
        get gene.txt (gene_id, seq_region_id, xref_id, canonical_transcript_id, stable_id)
          key: line[0] = gene_id
          tupple:
            - $3  = seq_region_id
            - $4  = start
            - $5  = end
            - $6  = strand
            - $7  = xref_id
            - $11 = canonical_transcript_id
            - $12 = stable_id (ENSG)
        '''

        '''
        GENES
        '''
        gene_url = f"{self.url}gene.txt.gz"
        gene_r = requests.get(gene_url)
        gene_str = gzip.decompress(gene_r.content).decode()
        gene_id = {}
        gene_name = {}
        xrefs = {}
        # ~ print("seq_region_ids:", self.seq_region_ids)
        for line in gene_str.splitlines():
            line = line.split('\t')
            # ~ print("line:", line)
            # ~ if line[3] in self.seq_region_ids:   # item must be in a regular chromosom
            if line[3] in self.chr_dict:   # item must be in a regular chromosom
                gene_id[line[0]] = (line[3], line[7], line[11], line[12])
                # ~ print('line3:', line[3])
                # ~ print(f"self.seq_region_ids is a {type(self.seq_region_ids)}")
                gene_name[line[12]] = {'chr': self.chr_dict[line[3]],
                                       'start': int(line[4]),
                                       'end': int(line[5]),
                                       'strand': int(line[6]),
                                       'biotype': line[1],
                                       'desc': line[9],
                                       }
                xrefs.setdefault(line[7], []).append(line[0])

        del gene_r, gene_str

        self.refs["gene"] = gene_id
        self.refs["xref"] = xrefs
        self.data["gene"] = gene_name


    def _add_symbols(self):
        '''
        XREF - list of symbol-name
        kept  if xref_id is found in refs[xrefs] dict
          - add symbol names in data[gene] dict
          - populate data[symbol] dict (symbol-name: [ENSG01, ENSG02])
        '''
        if __name__ == "__main__": print("Get xrefs.txt, please wait...")

        xref_url = f"{self.url}xref.txt.gz"
        xref_r = requests.get(xref_url)
        xref_str = gzip.decompress(xref_r.content).decode()
        for line in xref_str.splitlines():
            line = line.split()
            if line[0] in self.refs["xref"]:
                for gene_id in self.refs["xref"][line[0]]:
                    ensg = self.refs["gene"][gene_id][3]
                    xref_id = line[0]
                    symbol = line[3]
                    self.data["gene"][ensg]["symbol"] = symbol                        # add to gene info
                    self.data["symbol"].setdefault(symbol.upper(), []).append(ensg)   # add to symbol index
        del xref_r, xref_str


    def _add_aliases(self):
        '''
        SYNONYM - list of aliases (from external sources)
        kept if xref_id is found on refs[xref] dict
          - add alias in data[gene] dict
          - populate data[alias] dict (alias: [symbol1, symbol2])
        '''
        if __name__ == "__main__": print("Get external_synonym.txt, please wait...")

        ### EXT_SYNONYMS - get external_synonym.txt (Aliases)
        ext_synonym_url = f"{self.url}external_synonym.txt.gz"
        ext_synonym_r = requests.get(ext_synonym_url)
        ext_synonyms_str = gzip.decompress(ext_synonym_r.content).decode()
        aliases = {}
        for line in ext_synonyms_str.splitlines():
            line = line.split('\t')
            if line[0] in self.refs["xref"]:
                alias = line[1]
                for gene_id in self.refs["xref"][line[0]]:
                    ensg = self.refs["gene"][gene_id][3]
                    symbol = self.data["gene"][ensg]["symbol"]
                    self.data["gene"][ensg].setdefault('aliases', []).append(alias)           # add to gene info
                    aliases.setdefault(alias.upper(), set()).add(ensg)                        # add to alias index
        self.data["alias"] = { k:list(v) for k,v in aliases.items()}


    def _add_transcripts(self):
        '''
        TRANSCRIPT - list of transcripts
        $0 = transcript_id
        $1 = gene_id
        $3 = region_id
        $13 = stable_id (ENST)
        '''
        if __name__ == "__main__": print("Get transcript.txt, please wait...")

        transcript_url = f"{self.url}transcript.txt.gz"
        transcript_r = requests.get(transcript_url)
        transcript_str = gzip.decompress(transcript_r.content).decode()
        for line in transcript_str.splitlines():
            line = line.split('\t')
            if line[1] in self.refs["gene"]:
                ensg_id = line[1]
                ensg = self.refs["gene"][ensg_id][3]
                enst = line[13]
                self.data["gene"][ensg].setdefault("transcript", []).append(enst)  # add to gene info
                self.data["transcript"][enst.upper()] = ensg                       # add to transcript index
                if line[0] == self.refs["gene"][ensg_id][2]:
                    self.data["gene"][ensg]["canonical"] = enst                    # add to gene info if canonical


    def _write(self):
        if "\\N" in self.data["symbol"]:
            del self.data["symbol"]["\\N"]

        # ~ with open(self.outfile, 'w', encoding='utf8') as outfile:
            # ~ yaml.dump(self.data, outfile, default_flow_style=False, allow_unicode=True)
        with open(self.outfile, 'wb') as fh:
            pickle.dump(self.data, fh)
            if self.args.debug: print(f"{DEBUG}File {os.path.basename(self.outfile)!r} created.{ENDCOL}")
        return


    def _to_report(self):
        """ Function doc """
        self.report.append("\n## Gene Info\n")
        self.report.append(f"- nb genes: {len(self.data['gene'])}")
        self.report.append(f"- nb symbols: {len(self.data['symbol'])}")
        self.report.append(f"- nb aliases: {len(self.data['alias'])}")
        self.report.append(f"- nb transcripts: {len(self.data['transcript'])}")
